"""Useful functions for parsing"""

from pathlib import Path
from .types import Alpha, Generation, GraphName


def build_path_to_oplib_instance(
    oplib_root: Path,
    generation: Generation,
    name: GraphName,
    alpha: Alpha = Alpha.fifty,
) -> Path:
    """Build a filepath to a oplib instance"""
    filename: str = name + "-" + generation.value + "-" + str(alpha.value) + ".oplib"
    return oplib_root / "instances" / generation.value / filename


def build_path_to_tsplib_instance(tsplib_root: Path, name: GraphName) -> Path:
    """Build a filepath to a tsplib instance"""
    filename = name.value + ".tsp"
    return tsplib_root / filename



"""Travelling Salesman Problem with Profits library"""

from .converter import (
    asymmetric_from_directed,
    asymmetric_from_undirected,
    biggest_vertex_id_from_graph,
    get_original_from_split_vertex,
    get_original_path_from_split_path,
    head_prize,
    is_split_vertex_pair,
    is_vertex_split_head,
    is_vertex_split_tail,
    split_head,
    split_tail,
    tail_prize,
    to_vertex_dataframe,
)
from .complete import is_complete, is_complete_with_self_loops
from .problem import ProfitsProblem
from .utils import build_path_to_oplib_instance, build_path_to_tsplib_instance
from .types import (
    Alpha,
    DisjointPaths,
    Edge,
    EdgeFunction,
    EdgeFunctionName,
    EdgeList,
    Generation,
    GraphName,
    OptimalSolutionTSP,
    Vertex,
    VertexFunction,
    VertexFunctionName,
    VertexList,
    VertexLookup,
)
from .walk import (
    edge_list_from_walk,
    is_simple_cycle,
    is_simple_path,
    is_walk,
    total_cost,
    total_cost_graph_tool,
    total_cost_networkx,
    total_prize,
)

__all__ = [
    "Alpha",
    "DisjointPaths",
    "Edge",
    "EdgeFunction",
    "EdgeFunctionName",
    "EdgeList",
    "Generation",
    "GraphName",
    "OptimalSolutionTSP",
    "ProfitsProblem",
    "Vertex",
    "VertexFunction",
    "VertexFunctionName",
    "VertexList",
    "VertexLookup",
    "asymmetric_from_directed",
    "asymmetric_from_undirected",
    "biggest_vertex_id_from_graph",
    "build_path_to_oplib_instance",
    "build_path_to_tsplib_instance",
    "complete",
    "converter",
    "edge_list_from_walk",
    "get_original_from_split_vertex",
    "get_original_path_from_split_path",
    "head_prize",
    "is_complete",
    "is_complete_with_self_loops",
    "is_simple_cycle",
    "is_simple_path",
    "is_split_vertex_pair",
    "is_vertex_split_head",
    "is_vertex_split_tail",
    "is_walk",
    "problem",
    "split_head",
    "split_tail",
    "tail_prize",
    "to_vertex_dataframe",
    "total_cost",
    "total_cost_graph_tool",
    "total_cost_networkx",
    "total_prize",
    "types",
    "utils",
    "walk",
]



"""Test converting asymmetric graphs to / from directed & undirected graphs"""

import pytest
import networkx as nx
from tspwplib import EdgeFunctionName, VertexFunctionName
from tspwplib.converter import (
    asymmetric_from_directed,
    asymmetric_from_undirected,
    biggest_vertex_id_from_graph,
    get_original_from_split_vertex,
    get_original_path_from_split_path,
    head_prize,
    is_split_vertex_pair,
    is_vertex_split_head,
    is_vertex_split_tail,
    split_head,
    split_tail,
    tail_prize,
)


def original_and_asymmetric_comparison(
    original_graph: nx.Graph, asymmetric_graph: nx.DiGraph
) -> None:
    """Compare original and asymmetric graph"""
    biggest_vertex = biggest_vertex_id_from_graph(original_graph)
    for vertex, vertex_data in original_graph.nodes(data=True):
        head = split_head(biggest_vertex, vertex)
        tail = split_tail(biggest_vertex, vertex)
        assert asymmetric_graph.has_node(head)
        assert asymmetric_graph.has_node(tail)
        assert asymmetric_graph.has_edge(tail, head)
        assert len(list(asymmetric_graph.predecessors(head))) == 1
        assert len(list(asymmetric_graph.successors(tail))) == 1
        # check node data
        head_data = asymmetric_graph.nodes(data=True)[head]
        tail_data = asymmetric_graph.nodes(data=True)[tail]
        assert vertex_data.keys() == head_data.keys()
        assert vertex_data.keys() == tail_data.keys()
        assert (
            vertex_data[VertexFunctionName.prize]
            == head_data[VertexFunctionName.prize] + tail_data[VertexFunctionName.prize]
        )
        # check edge data
        edge_data = asymmetric_graph[tail][head]
        assert edge_data[EdgeFunctionName.cost] == 0

    for u, v, original_edge_data in original_graph.edges(data=True):
        head_of_u = split_head(biggest_vertex, u)
        tail_of_v = split_tail(biggest_vertex, v)
        assert asymmetric_graph.has_edge(head_of_u, tail_of_v)
        # check edge data
        edge_data = asymmetric_graph[head_of_u][tail_of_v]
        for key, value in original_edge_data.items():
            assert key in edge_data
            assert edge_data[key] == value


def test_asymmetric_from_directed(directed_complete_graph):
    """Test asymmetric graphs can be created from directed graphs"""
    asymmetric_graph = asymmetric_from_directed(directed_complete_graph)
    original_and_asymmetric_comparison(directed_complete_graph, asymmetric_graph)
    assert nx.is_strongly_connected(asymmetric_graph)


def test_asymmetric_from_undirected(undirected_complete_graph):
    """Test asymmetric graphs can be created from undirected graphs"""
    asymmetric_graph = asymmetric_from_undirected(undirected_complete_graph)
    original_and_asymmetric_comparison(undirected_complete_graph, asymmetric_graph)
    assert nx.is_strongly_connected(asymmetric_graph)


def test_biggest_vertex_id_from_graph(undirected_complete_graph):
    """Test the biggest vertex id is returned"""
    for u in undirected_complete_graph:
        assert u <= biggest_vertex_id_from_graph(undirected_complete_graph)


def test_get_original_from_split_vertex(
    original_vertices, split_head_vertices, split_tail_vertices
):
    """Test we can return the original vertex"""
    biggest_vertex = max(original_vertices)
    for i, original in enumerate(original_vertices):
        assert original == get_original_from_split_vertex(
            biggest_vertex, split_head_vertices[i]
        )
        assert original == get_original_from_split_vertex(
            biggest_vertex, split_tail_vertices[i]
        )


@pytest.mark.parametrize(
    "split_path,original_path",
    [
        ([3, 6, 4, 7, 5, 8], [0, 1, 2]),
        ([6, 4, 7, 5], [0, 1, 2]),
        ([], []),
        ([8, 3], [2, 0]),
        ([5, 8, 3, 6, 5], [2, 0, 2]),
    ],
)
def test_get_original_path_from_split_path(
    original_vertices,
    split_path,
    original_path,
):
    """Test we get the path from the original graph given split path"""
    biggest_vertex = max(original_vertices)
    assert original_path == get_original_path_from_split_path(
        biggest_vertex, split_path
    )


def test_is_split_vertex_pair(
    original_vertices, split_head_vertices, split_tail_vertices
):
    """Test if a pair of split vertices are a pair"""
    biggest_vertex = max(original_vertices)
    for original, head, tail in zip(
        original_vertices, split_head_vertices, split_tail_vertices
    ):
        assert is_split_vertex_pair(biggest_vertex, tail, head)
        assert not is_split_vertex_pair(biggest_vertex, tail, original)


def test_is_vertex_split_head(
    original_vertices, split_head_vertices, split_tail_vertices
):
    """Test split vertices are correctly identified as heads"""
    biggest_vertex = max(original_vertices)
    for original, head, tail in zip(
        original_vertices, split_head_vertices, split_tail_vertices
    ):
        assert is_vertex_split_head(biggest_vertex, head)
        assert not is_vertex_split_head(biggest_vertex, original)
        assert not is_vertex_split_head(biggest_vertex, tail)


def test_is_vertex_split_tail(
    original_vertices, split_head_vertices, split_tail_vertices
):
    """Test split vertices are correctly identfied as tails"""
    biggest_vertex = max(original_vertices)
    for original, head, tail in zip(
        original_vertices, split_head_vertices, split_tail_vertices
    ):
        assert not is_vertex_split_tail(biggest_vertex, head)
        assert not is_vertex_split_tail(biggest_vertex, original)
        assert is_vertex_split_tail(biggest_vertex, tail)


def test_split_head(original_vertices, split_head_vertices):
    """Test we split head correctly"""
    biggest_vertex = max(original_vertices)
    for original, head in zip(original_vertices, split_head_vertices):
        assert split_head(biggest_vertex, original) == head


def test_split_tail(original_vertices, split_tail_vertices):
    """Test we split tails correctly"""
    biggest_vertex = max(original_vertices)
    for original, tail in zip(original_vertices, split_tail_vertices):
        assert split_tail(biggest_vertex, original) == tail


@pytest.mark.parametrize("prize, prize_of_tail", [(8, 4), (7, 3), (0, 0)])
def test_tail_prize(prize, prize_of_tail):
    """Test prizes of tail are split correctly"""
    assert isinstance(tail_prize(prize), int)
    assert tail_prize(prize) == prize_of_tail


@pytest.mark.parametrize("prize, prize_of_head", [(8, 4), (7, 4), (0, 0)])
def test_head_prize(prize, prize_of_head):
    """Test prizes of head are split correctly"""
    assert isinstance(head_prize(prize), int)
    assert head_prize(prize) == prize_of_head



"""Functions for walks in a graph"""

from typing import Mapping

import graph_tool as gt
import networkx as nx
from .types import Edge, EdgeFunctionName, EdgeList, Vertex, VertexList


def edge_list_from_walk(walk: VertexList) -> EdgeList:
    """Get ordered list of edges from an ordered list of vertices

    Args:
        walk: Ordered list of vertices that represent a walk in the graph

    Returns:
        List of edges in the same order as the walk
    """
    edge_list: EdgeList = []
    if len(walk) <= 1:
        return edge_list
    for i in range(len(walk) - 1):
        vertex = walk[i]
        next_vertex = walk[i + 1]
        edge_list.append((vertex, next_vertex))
    return edge_list


def is_walk(G: nx.Graph, walk: VertexList) -> bool:
    """Is the walk a sequence of adjacent vertices in the graph?

    Args:
        G: input graph
        walk: Ordered sequence of vertices

    Returns:
        True if all vertices are adjacent in the graph
    """
    edge_list = edge_list_from_walk(walk)
    return all([G.has_edge(u, v) for u, v in edge_list])


def is_simple_cycle(G: nx.Graph, cycle: VertexList) -> bool:
    """Is the cycle simple in the graph?

    Args:
        G: input graph
        cycle: Ordered sequence of vertices

    Returns:
        True if the cycle is simple in the graph
    """
    cycle_length = len(cycle)
    if cycle_length <= 1:
        return True
    return (
        is_walk(G, cycle)
        and cycle_length == len(set(cycle)) + 1
        and cycle[0] == cycle[cycle_length - 1]
    )


def is_simple_path(G: nx.Graph, path: VertexList) -> bool:
    """Is the path simple in the graph?

    Args:
        G: input graph
        path: Ordered sequence of vertices

    Returns:
        True if the path is simple in the graph
    """
    return is_walk(G, path) and len(path) == len(set(path))


def total_prize(prizes: Mapping[Vertex, int], vertices: VertexList) -> int:
    """Total prize of vertices

    Args:
        prizes: A mapping from vertices to prizes, e.g. dict, property map
        vertices: List of vertices in the prizes map

    Returns:
        Total prize of vertices
    """
    sum_prize: int = 0
    for vertex in vertices:
        sum_prize += prizes[vertex]
    return sum_prize


def total_cost(costs: Mapping[Edge, int], edges: EdgeList) -> int:
    """Total cost of edges

    Args:
        costs: Mapping from edges to costs
        edges: List of edges

    Returns:
        Total cost of edges
    """
    sum_cost: int = 0
    for edge in edges:
        try:
            sum_cost += costs[edge]
        except KeyError:
            try:
                u, v = edge
                sum_cost += costs[(v, u)]
            except KeyError as second_key_error:
                raise KeyError(
                    "Edge ({u},{v}) or ({v},{u}) do not exist in costs map".format(
                        u=u, v=v
                    )
                ) from second_key_error
        except Exception as error:
            raise KeyError(
                "{edge} does not exist in cost map".format(edge=edge)
            ) from error
    return sum_cost


def total_cost_graph_tool(graph: gt.Graph, walk: VertexList) -> int:
    """Get the total cost of edges in a walk of the graph-tool graph

    Args:
        graph: Undirected input graph
        walk: A sequence of adjacent vertices

    Returns:
        Total cost of edges in walk
    """
    edges_in_tour = edge_list_from_walk(walk)
    gt_edges = [graph.edge(*e) for e in edges_in_tour]
    return total_cost(graph.ep.cost, gt_edges)


def total_cost_networkx(graph: nx.Graph, walk: VertexList) -> int:
    """Get the total cost of edges in a walk of the networkx graph

    Args:
        graph: Undirected input graph with cost attribute
        walk: A sequence of adjacent vertices

    Returns:
        Total cost of edges in the walk
    """
    edges_in_tour = edge_list_from_walk(walk)
    cost_attr = nx.get_edge_attributes(graph, EdgeFunctionName.cost.value)
    return total_cost(cost_attr, edges_in_tour)



"""Fixtures for testing converting graphs"""

import networkx as nx
import pytest
from tspwplib import EdgeFunctionName, VertexFunctionName, VertexList


@pytest.fixture(scope="function", params=[3, 5, 10])
def undirected_complete_graph(request) -> nx.Graph:
    """Undirected complete graph on 3, 5 & 10 vertices"""
    graph = nx.complete_graph(request.param)
    nx.set_node_attributes(graph, 1, name=VertexFunctionName.prize)
    nx.set_edge_attributes(graph, 1, name=EdgeFunctionName.cost)
    return graph


@pytest.fixture(scope="function", params=[3, 5, 10])
def directed_complete_graph(request) -> nx.DiGraph:
    """Undirected complete graph on 3, 5 & 10 vertices"""
    graph = nx.complete_graph(request.param, nx.DiGraph())
    nx.set_node_attributes(graph, 1, name=VertexFunctionName.prize)
    nx.set_edge_attributes(graph, 1, name=EdgeFunctionName.cost)
    return graph


@pytest.fixture(scope="function")
def original_vertices() -> VertexList:
    """List of vertices in the original graph"""
    return list(range(3))


@pytest.fixture(scope="function")
def split_head_vertices() -> VertexList:
    """Split head vertices"""
    return list(range(6, 9))


@pytest.fixture(scope="function")
def split_tail_vertices() -> VertexList:
    """Split tail vertices"""
    return list(range(3, 6))


@pytest.fixture(scope="function")
def split_vertices() -> VertexList:
    """List of vertices in asymmetric graph"""
    return list(range(3, 9))



"""Functions and classes for datasets"""

import random
from typing import List
import graph_tool as gt
import networkx as nx
import tsplib95
from .types import EdgeList, Vertex, VertexLookup


class ProfitsProblem(tsplib95.models.StandardProblem):
    """TSP with Profits Problem

    You can set `edge_removal_probability` to remove edges with this probability.
    """

    # Maximum distance of the total route in a OP.
    cost_limit = tsplib95.fields.IntegerField("COST_LIMIT")
    # The scores of the nodes of a OP are given in the form (per line)
    node_score = tsplib95.fields.DemandsField("NODE_SCORE_SECTION")
    # The optimal solution to the TSP
    tspsol = tsplib95.fields.IntegerField("TSPSOL")

    def __init__(
        self, edge_removal_probability: float = 0.0, seed: int = 0, special=None, **data
    ):
        super().__init__(special=special, **data)
        self._edge_removal_probability = edge_removal_probability
        self._seed = seed

    @property
    def edge_removal_probability(self) -> float:
        """Probability of removing an edge from the graph.

        It is strongly recommended to only set this value in the constructor.
        """
        return self._edge_removal_probability

    def __set_edge_attributes(self, graph: nx.Graph, names: VertexLookup) -> None:
        """Set edge attributes"""
        # add every edge with some associated metadata
        for u, v in self.get_edges():
            cost: int = self.get_weight(u, v)
            # pylint: disable=unsupported-membership-test
            # is_fixed: bool = (u, v) in self.fixed_edges
            graph.add_edge(names[u], names[v], cost=cost)

    def __set_graph_attributes(self, graph: nx.Graph) -> None:
        """Set attributes of the graph such as the name"""
        graph.graph["name"] = self.name
        graph.graph["comment"] = self.comment
        graph.graph["type"] = self.type
        graph.graph["dimension"] = self.dimension
        graph.graph["capacity"] = self.capacity
        graph.graph["root"] = self.get_root_vertex()

    def __set_node_attributes(self, graph: nx.Graph, names: VertexLookup) -> None:
        """Add node attributes"""
        node_score = self.get_node_score()
        for vertex in list(self.get_nodes()):
            # NOTE pyintergraph cannot handle bool, so we remove some attributes:
            # is_depot, demand, display
            # pylint: disable=unsupported-membership-test,no-member
            # is_depot = vertex in self.depots
            coord: List[int] = self.node_coords.get(vertex)
            graph.add_node(
                names[vertex],
                x=coord[0],
                y=coord[1],
                prize=node_score[vertex],
                # is_depot=is_depot,
            )
            # demand: int = self.demands.get(vertex)
            # display = self.display_data.get(vertex)
            # if not demand is None:
            # graph[vertex]["demand"] = demand
            # if not display is None:
            # graph[vertex]["display"] = display
        # nx.set_node_attributes(
        #     graph, self.get_node_score(), name=VertexFunctionName.prize
        # )

    def get_graph(self, normalize: bool = False) -> nx.Graph:
        """Return a networkx graph instance representing the problem.

        Args:
            normalize: rename nodes to be zero-indexed
        """
        # directed graphs are fundamentally different
        graph: nx.Graph = nx.Graph() if self.is_symmetric() else nx.DiGraph()

        # set up a map from original node name to new node name
        nodes: List[Vertex] = list(self.get_nodes())
        if normalize:
            names = {n: i for i, n in enumerate(nodes)}
        else:
            names = {n: n for n in nodes}

        self.__set_node_attributes(graph, names)
        self.__set_edge_attributes(graph, names)

        # add basic graph metadata
        self.__set_graph_attributes(graph)
        return graph

    def get_graph_tool(self, normalize: bool = True) -> gt.Graph:
        """Return a graph tools undirected graph

        Args:
            normalize: rename nodes to be zero-indexed
        """
        graph = gt.Graph(directed=not self.is_symmetric())

        # by default normalize because graph tools index starts at zero
        nodes: List[Vertex] = list(self.get_nodes())
        if normalize:
            names = {n: i for i, n in enumerate(nodes)}
        else:
            names = {n: n for n in nodes}
        graph.add_vertex(len(nodes))

        # create list of edges
        edges = []
        for u, v in self.get_edges():
            if u <= v or not self.is_symmetric():
                edges.append((names[u], names[v], self.get_weight(u, v)))

        # assign cost to edges
        cost_property = graph.new_edge_property("int")
        graph.add_edge_list(edges, eprops=[cost_property])
        graph.ep.cost = cost_property

        # assign prize to vertices
        prize_property = graph.new_vertex_property("int")
        node_score = self.get_node_score()
        prize_list = [node_score[v + 1] for v in graph.get_vertices()]
        prize_property.a = prize_list
        graph.vertex_properties.prize = prize_property

        return graph

    def get_total_prize(self) -> int:
        """Get the sum of prize over all vertices

        Returns:
            Total prize
        """
        # pylint: disable=no-member
        return sum([value for _, value in self.get_node_score().items()])

    def get_quota(self, alpha: int) -> int:
        """The quota is alpha percent of the total prize

        Args:
            alpha: Percent of the total prize

        Returns:
            quota
        """
        if alpha > 100:
            raise ValueError("Cannot have a percent over 100 for alpha")
        if alpha < 0:
            raise ValueError("Cannot have a negative percent for alpha")
        return int(float(alpha * self.get_total_prize()) / 100.0)

    def number_of_nodes(self) -> int:
        """Get the number of nodes in the problem

        Returns:
            Number of nodes in graph
        """
        return len(list(self.get_nodes()))

    def get_cost_limit(self) -> int:
        """Get the cost limit for a TSP with Profits problem

        Returns:
            Cost limit
        """
        return self.cost_limit

    def get_node_score(self) -> VertexLookup:
        """Get the node scores (profits)

        Returns:
            Mapping from node to node score (profit)
        """
        return self.node_score

    def get_tsp_optimal_value(self) -> int:
        """Get the value of the optimal solution to TSP

        Returns:
            TSP optimal value
        """
        return self.tspsol

    def get_root_vertex(self, normalize: bool = False) -> Vertex:
        """Get the root vertex

        Args:
            normalize: If true, vertices start at index 0

        Returns:
            The first depot in the list

        Raises:
            ValueError: If the list of depots is empty
        """
        nodes: List[Vertex] = list(self.get_nodes())
        if normalize:
            names = {n: i for i, n in enumerate(nodes)}
        else:
            names = {n: n for n in nodes}
        try:
            # pylint: disable=unsubscriptable-object
            return names[self.depots[0]]
        except KeyError as key_error:
            raise ValueError("The list of depots is empty") from key_error

    def get_edges(self) -> EdgeList:
        """Get a list of edges in the graph

        If the `edge_removal_probability` is set in the constructor,
        then edges will be randomly removed

        Returns:
            List of edges in the graph
        """
        edges: EdgeList = list(super().get_edges())
        edges_copy = edges.copy()
        random.seed(self._seed)
        for edge in edges:
            # do not remove self-loops
            if (
                random.random() < self.edge_removal_probability
                and edge[0] != edge[1]
                and edge in edges_copy
            ):
                if self.is_symmetric() and edge[0] < edge[1]:
                    # remove both (u,v) and (v,u) in undirected case
                    edges_copy.remove(edge)
                    edges_copy.remove((edge[1], edge[0]))
                else:
                    # remove just (u,v) in directed case
                    edges_copy.remove(edge)
        return edges_copy



"""Tests for walks"""

import pytest
import networkx as nx
from tspwplib import (
    VertexFunctionName,
    edge_list_from_walk,
    is_simple_cycle,
    is_simple_path,
    is_walk,
    total_cost_graph_tool,
    total_cost_networkx,
    total_prize,
)


def test_is_open_walk_is_walk(walk_graph, open_walk):
    """Test walks"""
    assert is_walk(walk_graph, open_walk)


def test_is_not_a_walk(walk_graph, not_a_walk):
    """Test invalid walks are detected"""
    assert not is_walk(walk_graph, not_a_walk)


def test_edge_list_from_walk():
    """Test edge list from any walk"""
    assert edge_list_from_walk([]) == []
    assert edge_list_from_walk([0]) == []
    assert edge_list_from_walk([0, 1, 3, 1, 2]) == [(0, 1), (1, 3), (3, 1), (1, 2)]


def test_is_simple_path(walk_graph, simple_path):
    """Test simple paths"""
    assert is_simple_path(walk_graph, simple_path)


def test_walk_is_not_simple_path(walk_graph, open_walk):
    """Test walks with repeated vertices are not simple paths"""
    if len(open_walk) > 1:
        assert not is_simple_path(walk_graph, open_walk)


def test_simple_cycle_is_not_simple_path(walk_graph, simple_cycle):
    """Test cycles are not simple paths"""
    if len(simple_cycle) > 1:
        assert not is_simple_path(walk_graph, simple_cycle)


def test_is_simple_cycle(walk_graph, simple_cycle):
    """Test cycles are simple"""
    assert is_simple_cycle(walk_graph, simple_cycle)


def test_path_is_not_simple_cycle(walk_graph, simple_path):
    """Test paths are not cycles"""
    if len(simple_path) > 1:
        assert not is_simple_cycle(walk_graph, simple_path)


def test_total_cost_networkx(weighted_walk_networkx_graph):
    """Test total cost"""
    assert total_cost_networkx(weighted_walk_networkx_graph, [0, 1, 2]) == 1 + 3
    assert total_cost_networkx(weighted_walk_networkx_graph, []) == 0
    assert (
        total_cost_networkx(weighted_walk_networkx_graph, [0, 1, 2, 1, 0])
        == 1 + 3 + 3 + 1
    )
    with pytest.raises(KeyError):
        total_cost_networkx(weighted_walk_networkx_graph, [0, 1, 2, 3])


def test_total_prize(weighted_walk_networkx_graph):
    """Test total prize"""
    prizes = nx.get_node_attributes(
        weighted_walk_networkx_graph, VertexFunctionName.prize
    )
    assert total_prize(prizes, [0, 1, 2]) == 3
    assert total_prize(prizes, []) == 0


def test_total_cost_graph_tool(weighted_walk_graph_tool):
    """Test total cost"""
    assert total_cost_graph_tool(weighted_walk_graph_tool, [0, 1, 2]) == 1 + 3
    assert total_cost_graph_tool(weighted_walk_graph_tool, []) == 0
    assert (
        total_cost_graph_tool(weighted_walk_graph_tool, [0, 1, 2, 1, 0])
        == 1 + 3 + 3 + 1
    )
    with pytest.raises(KeyError):
        total_cost_graph_tool(weighted_walk_graph_tool, [0, 1, 2, 3])



"""Tests for sparsity"""


from tspwplib import (
    ProfitsProblem,
    is_complete_with_self_loops,
    build_path_to_oplib_instance,
)


def test_remove_random_edges_from_graph(
    oplib_root, generation, graph_name, alpha, edge_removal_probability
):
    """Test the right number of edges are removed"""
    filepath = build_path_to_oplib_instance(oplib_root, generation, graph_name, alpha)
    problem = ProfitsProblem.load(
        filepath, edge_removal_probability=edge_removal_probability, seed=5
    )
    complete_problem = ProfitsProblem.load(filepath)
    complete_graph = complete_problem.get_graph()
    smaller_graph = problem.get_graph()
    assert problem.edge_removal_probability == edge_removal_probability

    # edge cases
    if edge_removal_probability == 0:
        assert smaller_graph.number_of_edges() == complete_graph.number_of_edges()
    elif edge_removal_probability == 1.0:
        assert smaller_graph.number_of_edges() == smaller_graph.number_of_nodes()
    # sufficient number of nodes for randomness to not have big effect
    elif smaller_graph.number_of_nodes() > 50:
        assert not is_complete_with_self_loops(smaller_graph)
        num_edges_not_self_loops = (
            complete_graph.number_of_edges() - complete_graph.number_of_nodes()
        )
        num_edges_lower_bound = float(num_edges_not_self_loops) * (
            1.0 - edge_removal_probability - 0.1
        )
        num_edges_upper_bound = float(num_edges_not_self_loops) * (
            1.0 - edge_removal_probability + 0.1
        )
        assert (
            num_edges_lower_bound
            <= smaller_graph.number_of_edges() - smaller_graph.number_of_nodes()
            <= num_edges_upper_bound
        )



"""Test the TSP with Profits Problem class"""

import math
import pytest
import networkx as nx
from tspwplib import (
    Generation,
    ProfitsProblem,
    EdgeFunctionName,
    OptimalSolutionTSP,
    VertexFunctionName,
    build_path_to_oplib_instance,
)


def test_parse_profits_problem(oplib_root, generation, graph_name, alpha):
    """Test an OP instance can be parsed"""
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    assert "COST_LIMIT" in ProfitsProblem.fields_by_keyword
    assert "NODE_SCORE_SECTION" in ProfitsProblem.fields_by_keyword

    problem = ProfitsProblem.load(filepath)
    assert problem.is_complete()


def test_get_cost_limit(oplib_root, generation, graph_name, alpha):
    """Test we can get the cost limit"""
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    expected_cost_limit = math.ceil(
        OptimalSolutionTSP[graph_name] * (alpha.value / 100.0)
    )
    assert problem.get_cost_limit() == expected_cost_limit


def test_get_node_score(oplib_root, generation, graph_name, alpha):
    """Test every node has a prize"""
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    node_scores = problem.get_node_score()
    assert isinstance(node_scores, dict)
    for vertex in problem.get_nodes():
        assert vertex in node_scores
        value = node_scores[vertex]
        assert value >= 0


def test_get_graph(oplib_root, generation, graph_name, alpha):
    """Test we can load a graph with prizes"""
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    graph = problem.get_graph()
    prizes = problem.get_node_score()

    assert nx.complete_graph(graph)

    for vertex in range(1, graph.number_of_nodes() + 1):
        assert vertex in graph

    for vertex, data in graph.nodes(data=True):
        assert data[VertexFunctionName.prize] == prizes[vertex]

    # check every edge has an attribute called 'cost'
    for _, _, data in graph.edges(data=True):
        assert EdgeFunctionName.cost in data

    # bool and list breaks pyintergraph so we avoid it
    valid_types = [str, int, float]
    for _, _, data in graph.edges(data=True):
        for _, value in data.items():
            assert type(value) in valid_types

    for _, data in graph.nodes(data=True):
        for _, value in data.items():
            assert type(value) in valid_types

    assert graph.graph["root"] == 1


def test_get_graph_tool(oplib_root, generation, graph_name, alpha):
    """Test returning graph tool undirected weighted graph"""
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    gt_graph = problem.get_graph_tool()
    nx_graph = problem.get_graph(normalize=True)
    assert nx_graph.has_node(0)
    assert 0 in gt_graph.get_vertices()
    assert gt_graph.num_vertices() == nx_graph.number_of_nodes()
    assert gt_graph.num_edges() == nx_graph.number_of_edges()

    # check cost
    for u, v, data in nx_graph.edges(data=True):
        gt_edge = gt_graph.edge(u, v, add_missing=False)
        assert gt_edge
        assert gt_graph.ep.cost[gt_edge] == data["cost"]
    # check prize on vertices
    for u, data in nx_graph.nodes(data=True):
        assert data["prize"] == gt_graph.vertex_properties.prize[u]


def test_get_root_vertex(oplib_root, generation, graph_name, alpha):
    """Test the root vertex is 1 when un-normalized (0 when normalized)"""
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    assert problem.get_root_vertex(normalize=False) == 1
    assert problem.get_root_vertex(normalize=True) == 0


def test_get_total_prize(oplib_root, graph_name, alpha):
    """Test total prize"""
    generation = Generation.gen1
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    assert problem.get_total_prize() == problem.number_of_nodes()


def test_get_quota(oplib_root, graph_name, alpha):
    """Test the quota is calculated properly"""
    generation = Generation.gen1
    filepath = build_path_to_oplib_instance(
        oplib_root, generation, graph_name, alpha=alpha
    )
    problem = ProfitsProblem.load(filepath)
    assert problem.get_quota(10) == int(0.1 * problem.number_of_nodes())
    assert problem.get_quota(50) == int(0.5 * problem.number_of_nodes())
    assert problem.get_quota(90) == int(0.9 * problem.number_of_nodes())
    assert problem.get_quota(100) == int(1.0 * problem.number_of_nodes())
    assert problem.get_quota(0) == int(0.0 * problem.number_of_nodes())
    with pytest.raises(ValueError):
        problem.get_quota(-1)
        problem.get_quota(101)
